/*-
 * Copyright (c) 2010 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "config.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/wait.h>

#include <arpa/inet.h>

#include <bzlib.h>
#include <ctype.h>
#include <inttypes.h>
#include <md5.h>
#include <netdb.h>
#include <regex.h>
#include <rmd160.h>
#include <sha1.h>
#include <sha2.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <vis.h>
#include <zlib.h>

#include "b64.h"
#include "hashes.h"
#include "codecs.h"

#ifndef USE_ARG
#define USE_ARG(x)	/*LINTED*/(void)&x
#endif

#ifndef __UNCONST
#define __UNCONST(a)	((void *)(unsigned long)(const void *)(a))
#endif

#ifndef __arraycount
#define	__arraycount(__x)	(sizeof(__x) / sizeof(__x[0]))
#endif

#define HEX_CHAR_SET	"0123456789abcdef"

/* add an error message to the codecs error buffer */
static int
add_error(codecs_t *codecs, const char *fmt, ...)
{
	va_list	args;
	int	cc;

	va_start(args, fmt);
	cc = vsnprintf(&codecs->errors[codecs->errc],
			CODECS_ERRBUF_SIZE - codecs->errc, fmt, args);
	va_end(args);
	if (cc < (int)(CODECS_ERRBUF_SIZE - codecs->errc) && cc > 0) {
		codecs->errc += cc;
	}
	return 0;
}

 /* convert a hex digit in ascii to a number [0-15] */
static int
hexchar2num(const char ch)
{
	const char	*hex = HEX_CHAR_SET;
	const char	*hp;

	return ((hp = strchr(hex, ch)) == NULL) ? 0 : (int)(hp - hex);
}

/* convert from percent-encoded URI to "ASCII" */
static int
uri2ascii(codecs_t *codecs, const char *in, const size_t len,
	const char *op, void *outvp, size_t size)
{
	const char	*inp;
	char		*out = outvp;
	char		*outp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (inp = in, outp = out ;
	     (size_t)(inp - in) < len && (size_t)(outp - out) < size;
	     ) {
		if (*inp == '%') {
			*outp++ = (hexchar2num(*(inp + 1)) * 16) +
					hexchar2num(*(inp + 2));
			inp += 3;
		} else {
			*outp++ = *inp++;
		}
	}
	*outp = 0x0;
	return (int)(outp - out);
}

static const char percents[] = "!*\"'();:@&=+$,/?%#[]";

/* convert from "ASCII" to percent-encoded URI */
static int
ascii2uri(codecs_t *codecs, const char *in, const size_t len, const char *op,
	void *outvp, size_t size)
{
	const char	*inp;
	char		*out = outvp;
	char		*outp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (inp = in, outp = out ;
	     (size_t)(inp - in) < len && (size_t)(outp - out) < size;
	     ) {
		if (strchr(percents, *inp) == NULL) {
			*outp++ = *inp++;
		} else {
			*outp++ = '%';
			*outp++ = HEX_CHAR_SET[((unsigned)(*inp & 0xf0) >> 4)];
			*outp++ = HEX_CHAR_SET[*inp & 0xf];
			inp += 1;
		}
	}
	*outp = 0x0;
	return (int)(outp - out);
}

/* convert to lower from mixed case */
static int
mixed2lower(codecs_t *codecs, const char *in, const size_t len, const char *op,
	void *outvp, size_t size)
{
	const char	*inp;
	char		*out = outvp;
	char		*outp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (inp = in, outp = out ;
	     (size_t)(inp - in) < len && (size_t)(outp - out) < size;
	     inp++) {
		*outp++ = tolower((uint8_t)*inp);
	}
	*outp = 0x0;
	return (int)(outp - out);
}

/* convert to upper from mixed case */
static int
mixed2upper(codecs_t *codecs, const char *in, const size_t len, const char *op,
	void *outvp, size_t size)
{
	const char	*inp;
	char		*out = outvp;
	char		*outp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (inp = in, outp = out ;
	     (size_t)(inp - in) < len && (size_t)(outp - out) < size;
	     inp++) {
		*outp++ = toupper((uint8_t)*inp);
	}
	*outp = 0x0;
	return (int)(outp - out);
}

/* convert to unix line endings */
static int
dos2unix(codecs_t *codecs, const char *in, const size_t len, const char *op,
	void *outvp, size_t size)
{
	const char	*inp;
	char		*out = outvp;
	char		*outp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (inp = in, outp = out ;
	     (size_t)(inp - in) < len && (size_t)(outp - out) < size;
	     ) {
		if (*inp == '\r' && *(inp + 1) == '\n') {
			inp++;
		} 
		*outp++ = *inp++;
	}
	*outp = 0x0;
	return (int)(outp - out);
}

/* convert to dos line endings */
static int
unix2dos(codecs_t *codecs, const char *in, const size_t len, const char *op,
	void *outvp, size_t size)
{
	const char	*inp;
	char		*out = outvp;
	char		*outp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (inp = in, outp = out ;
	     (size_t)(inp - in) < len && (size_t)(outp - out) < size;
	     ) {
		if (*inp == '\n') {
			*outp++ = '\r';
		} 
		*outp++ = *inp++;
	}
	*outp = 0x0;
	return (int)(outp - out);
}

#define HEXDUMP_LINELEN	16

#ifndef PRIsize
#define PRIsize	"z"
#endif

/* show hexadecimal/ascii dump */
static int 
hexdump(codecs_t *codecs, const char *in, const size_t len, const char *op,
	void *outvp, size_t size)
{
	size_t	 i;
	char	 line[HEXDUMP_LINELEN + 1];
	char	*out = (char *)outvp;
	int	 o;

	USE_ARG(codecs);
	USE_ARG(op);
	for (i = 0, o = 0 ; i < len ; i++) {
		if (i % HEXDUMP_LINELEN == 0) {
			o += snprintf(&out[o], size - o,
					"%.5" PRIsize "u |  ", i);
		}
		o += snprintf(&out[o], size - o, "%.02x ", (uint8_t)in[i]);
		line[i % HEXDUMP_LINELEN] =
			(isprint((uint8_t)in[i])) ? in[i] : '.';
		if (i % HEXDUMP_LINELEN == HEXDUMP_LINELEN - 1) {
			line[HEXDUMP_LINELEN] = 0x0;
			o += snprintf(&out[o], size - o, " | %s\n", line);
		}
	}
	if (i % HEXDUMP_LINELEN != 0) {
		for ( ; i % HEXDUMP_LINELEN != 0 ; i++) {
			o += snprintf(&out[o], size - o, "   ");
			line[i % HEXDUMP_LINELEN] = ' ';
		}
		line[HEXDUMP_LINELEN] = 0x0;
		o += snprintf(&out[o], size - o, " | %s\n", line);
	}
	return (int)o;
}

/* convert from hexdump format to binary */
static int 
unhexdump(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	uint8_t	*out = (uint8_t *)outvp;
	size_t 	 o;
	size_t	 i;
	int	 done;
	int	 num;

	USE_ARG(codecs);
	USE_ARG(op);
	for (i = 0, o = 0, done = 0 ; !done && i < insize && o < outsize ; ) {
		i += 9;
		for (num = 0 ; !done && num < 16 && o < outsize ; num++) {
			if (!isxdigit((uint8_t)in[i])) {
				done = 1;
				break;
			}
			out[o++] = (hexchar2num(in[i]) * 16) +
					hexchar2num(in[i + 1]);
			i += 3;
		}
		i += 20;
	}
	return (int)o;
}

/* convert to visible string format */
static int
xlate_strvis_def(codecs_t *codecs, const char *in, const size_t len,
	const char *op, void *out, size_t size)
{
	USE_ARG(codecs);
	USE_ARG(op);
	if (size < len * 3) {
		return -1;
	}
	return strvisx(out, in, len, 0);
}

/* convert from visible string format */
static int
xlate_strunvis_def(codecs_t *codecs, const char *in, const size_t len,
	const char *op, void *out, size_t size)
{
	USE_ARG(codecs);
	USE_ARG(op);
	if (size < len / 3) {
		return -1;
	}
	return strunvisx(out, in, 0);
}

/* convert to visible string format C format */
static int
xlate_strvis_c(codecs_t *codecs, const char *in, const size_t len,
	const char *op, void *out, size_t size)
{
	USE_ARG(codecs);
	USE_ARG(op);
	if (size < len * 3) {
		return -1;
	}
	return strvisx(out, in, len, VIS_CSTYLE);
}

/* convert from visible string format C format */
static int
xlate_strunvis_c(codecs_t *codecs, const char *in, const size_t len,
	const char *op, void *out, size_t size)
{
	USE_ARG(codecs);
	USE_ARG(op);
	if (size < len / 3) {
		return -1;
	}
	return strunvisx(out, in, 0);
}

#define B64_LINELEN	80

/* wrapper around base64 encoding */
static int
xlate_b64encode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *out, size_t size)
{
	USE_ARG(codecs);
	USE_ARG(op);
	return b64encode(in, insize, out, size, B64_LINELEN - 8);
}

/* wrapper around base64 decoding */
static int
xlate_b64decode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *out, size_t size)
{
	USE_ARG(codecs);
	USE_ARG(op);
	return b64decode(in, insize, out, size);
}

/* uuenode/uudecode char set */
static const char uuchars[] =
	"`!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_";

/* uuencoded line size - # of input (binary) characters */
#define UU_LINE_SIZE	45

/* given a number in the range 0 to 0x3f, return the uu character */
static uint8_t
uuenc_char(char ch)
{
	return uuchars[(unsigned)ch & 0x3f];
}

/* given a uu character, return a number in the range 0 to 0x3f */
static int
uudec_char(uint8_t uch)
{
	const char	*cp;

	if ((cp = strchr(uuchars, uch)) != NULL) {
		return (uint8_t)(cp - uuchars);
	}
	return 0;
}

/* uuencoding - probably redundant now we have base64 */
static int
uuencode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	const char	*inp;
	uint32_t	 tmp;
	uint8_t		*out = outvp;
	uint8_t		*outp;
	int		 i;
	int		 n;

	USE_ARG(codecs);
	USE_ARG(op);
	for (outp = out, inp = in ;
	     (size_t)(inp - in) < insize  && (size_t)(outp - out) < outsize;
	     ) {
		/* 1 (up to) UU_LINE_SIZE (45) characters per line */ 
		n = MIN(UU_LINE_SIZE, (int)(insize - (int)(inp - in)));
		*outp++ = uuenc_char(n);
		for (i = 0; i < n ; i += 3, inp += 3) {
			tmp = ((inp[0] & 0xff) << 16);
			if ((size_t)(inp - in) + 1 < insize) {
				tmp |= ((inp[1] & 0xff) << 8);
			}
			if ((size_t)(inp - in) + 2 < insize) {
				tmp |= (inp[2] & 0xff);
			}
			*outp++ = uuenc_char((tmp >> 18) & 0x3f);
			*outp++ = uuenc_char((tmp >> 12) & 0x3f);
			*outp++ = uuenc_char((tmp >> 6) & 0x3f);
			*outp++ = uuenc_char((tmp) & 0x3f);
		}
		*outp++ = '\n';
	}
	return (int)(outp - out);
}

#define UU_HEADER	"begin "
#define UU_TRAILER	"`\nend\n"

/* full uuencoding - add the file header and tail */
static int
full_uuencode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	char	*out = outvp;
	int	 headc;
	int	 bodyc;
	int	 tailc;

	headc = snprintf(out, outsize, "%s644 uuencoded\n", UU_HEADER); 
	bodyc = uuencode(codecs, in, insize, op, &out[headc], outsize - headc);
	if (bodyc < 0) {
		return bodyc;
	}
	tailc = snprintf(&out[headc + bodyc], outsize - headc - bodyc,
		UU_TRAILER);
	return headc + bodyc + tailc;
}

/* uudecoding - probably redundant now we have base64 */
static int
uudecode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	const char	*inp;
	uint32_t	 tmp;
	size_t		 incc;
	char		*out = outvp;
	char		*outp;
	int		 i;
	int		 n;

	USE_ARG(codecs);
	USE_ARG(op);
	inp = in;
	incc = insize;
	if (strncmp(in, UU_HEADER, strlen(UU_HEADER)) == 0) {
		/* we have a uuencode header and trailer - modify sizes */
		for ( ; (size_t)(inp - in) < insize && *inp != '\n' ; inp++) {
		}
		if (*inp == '\n') {
			inp += 1;
		}
		incc -= (size_t)(inp - in) - strlen(UU_TRAILER);
	}
	for (outp = out ;
	     (size_t)(inp - in) < incc  && (size_t)(outp - out) < outsize;
	     ) {
		n = uudec_char(*inp++);
		for (i = 0; i < n ; i += 3, inp += 4) {
			tmp = ((uudec_char(inp[0]) & 0x3f) << 18) |
				((uudec_char(inp[1]) & 0x3f) << 12) |
				((uudec_char(inp[2]) & 0x3f) << 6) |
				(uudec_char(inp[3]) & 0x3f);
			if (i < n) {
				*outp++ = (tmp >> 16) & 0xff;
			}
			if (i + 1 < n) {
				*outp++ = (tmp >> 8) & 0xff;
			}
			if (i + 2 < n) {
				*outp++ = tmp & 0xff;
			}
		}
		if (*inp == '\n') {
			inp += 1;
		}
	}
	return (int)(outp - out);
}

/* translate fortran carriage control characters */
static int
asa(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	const char	*inp;
	char		*out = outvp;
	char		*outp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (inp = in, outp = out ;
	     (size_t)(inp - in) < insize && (size_t)(outp - out) < outsize;
	     ) {
		switch (*inp) {
		case '\n':
			switch(*(inp + 1)) {
			case ' ':
				*outp++ = *inp;
				inp += 2;
				break;
			case '0':
				*outp++ = '\n';
				inp += 2;
				break;
			case '1':
				*outp++ = '\f';
				inp += 2;
				break;
			case '+':
				*outp++ = '\r';
				inp += 2;
				break;
			default:
				*outp++ = *inp++;
				break;
			}
			break;
		default:
			*outp++ = *inp++;
			break;
		}
	}
	*outp = 0x0;
	return (int)(outp - out);
}

/* get a number from a substring */
static int
re2num(const char *s, int64_t from, int64_t to)
{
	int64_t	i;
	size_t	ret;
	int	sign;

	if ((i = from) < 0) {
		return 0;
	}
	sign = (s[(int)i] == '-') ? -1 : 1;
	for (ret = 0 ; i < to ; i++) {
		if (isdigit((uint8_t)s[(int)i])) {
			ret = (ret * 10) + (s[(int)i] - '0');
		}
	}
	return sign * ret;
}

/* get a substring of the current string */
static int
substring(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	regmatch_t	 matchv[10];
	regex_t	 	 r;
	char		*out = outvp;
	char		*outp;
	int		 from;
	int		 to;
	int		 i;

	/* XXX - find in translations table */
	USE_ARG(codecs);
	(void) regcomp(&r, "^substring:([-]?[0-9]+):([-]?[0-9]+)", REG_EXTENDED);
	(void) regexec(&r, op, 10, matchv, 0);
	if ((from = re2num(op, matchv[1].rm_so, matchv[1].rm_eo)) < 0) {
		/* makes no sense */
		from = 0;
	}
	if ((to = re2num(op, matchv[2].rm_so, matchv[2].rm_eo)) < 0) {
		/* adapt to length of input string */
		to = (int)insize;
	}
	for (outp = out, i = from ;
	     (size_t)(outp - out) < outsize && i < to;
	     i++) {
		*outp++ = in[i];
	}
	*outp = 0x0;
	return (int)(outp - out);
}

/* the base 85 functions are taken from:
	http://www.stillhq.com/svn/trunk/ascii85/ */
/* the source code is in the public domain */

/* utility routine to output a base 85 character */
static void
b85_write_char(char **outp, char **linep, const uint8_t ch)
{
	*(*outp)++ = ch;
	if ((int)(*outp - *linep) >= B64_LINELEN - 8) {
		*(*outp)++ = '\n';
		*linep = *outp;
	}
}

/* output code for base85 encoding */
static void
b85enc_out(char **outp, char **linep, unsigned tmp, int n)
{
	char	 buf[5];
	char	*bufp;
	int	 i;

	for (i = 0, bufp = buf ; i < 5 ; i++) {
		*bufp++ = (tmp % 85);
		tmp /= 85;
	}
	for (i = n ; i > 0 ; --i) {
		b85_write_char(outp, linep, *--bufp + '!');
	}
}

/* base85 encoding - 1.25 growth ratio, 'z' compression support, used in PDF */
static int
base85encode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	const char	*inp;
	uint32_t	 tmp;
	char		*out = outvp;
	char		*linep;
	char		*outp;
	int		 nbytes;

	USE_ARG(codecs);
	USE_ARG(op);
	linep = outp = out;
	*outp++ = '<';
	*outp++ = '~';
	tmp = 0; nbytes = 0;
	for (inp = in ;
	     (size_t)(inp - in) <= insize && (size_t)(outp - out) < outsize - 2;
	     ) {
		switch(nbytes++) {
		case 0:
			tmp |= ((*inp++ & 0xff) << 24);
			break;
		case 1:
			tmp |= ((*inp++ & 0xff) << 16);
			break;
		case 2:
			tmp |= ((*inp++ & 0xff) << 8);
			break;
		default:
			tmp |= (*inp++ & 0xff);
			b85enc_out(&outp, &linep, tmp, 5);
			tmp = 0;
			nbytes = 0;
			break;
		}
	}
	if (nbytes > 0) {
		b85enc_out(&outp, &linep, tmp, nbytes);
	}
	b85_write_char(&outp, &linep, '~');
	b85_write_char(&outp, &linep, '>');
	b85_write_char(&outp, &linep, '\n');
	return (int)(outp - out);
}

/* output char for char decode */
static void
b85dec_char(uint8_t **outp, uint32_t tmp, int count)
{
	switch (count) {
	case 4:
		*(*outp)++ = (tmp >> 24) & 0xff;
		*(*outp)++ = (tmp >> 16) & 0xff;
		*(*outp)++ = (tmp >> 8) & 0xff;
		*(*outp)++ = tmp & 0xff;
		break;
	case 3:
		*(*outp)++ = (tmp >> 24) & 0xff;
		*(*outp)++ = (tmp >> 16) & 0xff;
		*(*outp)++ = (tmp >> 8) & 0xff;
		break;
	case 2:
		*(*outp)++ = (tmp >> 24) & 0xff;
		*(*outp)++ = (tmp >> 16) & 0xff;
		break;
	case 1:
		*(*outp)++ = (tmp >> 24) & 0xff;
		break;
	}
}

/* base85 decoding */
static int
base85decode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	static const unsigned	 pow85[] = { 85*85*85*85, 85*85*85, 85*85, 85, 1 };
	const char		*inp;
	uint32_t		 tmp;
	uint8_t			 ch;
	uint8_t			*out = outvp;
	uint8_t			*outp;
	int			 nbytes;

	USE_ARG(codecs);
	USE_ARG(op);
	tmp = 0;
	nbytes = 0;
	for (inp = &in[2], outp = out ;
	     (size_t)(inp - in) <= insize && (size_t)(outp - out) < outsize;
	     ) {
		switch(ch = *inp++) {
		case '\n':
			break;
		case 'z':
			b85dec_char(&outp, 0, 4);
			break;
		case '~':
			if (*inp == '\n') {
				inp += 1;
			}
			if (*inp == '>') {
				if (nbytes == 0) {
					outp -= 1;
				} else {
					tmp += pow85[nbytes - 1];
					b85dec_char(&outp, tmp, nbytes - 1);
				}
				return (int)(outp - out);
			}
			break;
		default:
			tmp += (ch - '!') * pow85[nbytes++];
			if (nbytes == 5) {
				b85dec_char(&outp, tmp, 4);
				nbytes = 0;
				tmp = 0;
			}
			break;
		}
	}
	return (int)(outp - out);
}

/* perform sed commands on the input */
static int
sed(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	regmatch_t	 matchv[10];
	const char	*sep;
	const char	*inp;
	regex_t		 r;
	char		 pat[512];
	char		*out = (char *)outvp;
	char		*outp;
	char		*from;
	char		*to;
	int		 fromlen;
	int		 tolen;

	USE_ARG(codecs);
	(void) memset(&r, 0x0, sizeof(r));
	for (sep = &op[3] ; *sep && *sep != ':' ; sep++) {
	}
	/* jump over 's' - next non-whitespace is the separator */
	for (sep += 2; *sep && (*sep == ' ' || *sep == '\t') ; sep++) {
	}
	if (*sep == 0x0) {
		(void) fprintf(stderr, "bad regexp '%s'\n", op);
		return 0;
	}
	if (sep[strlen(sep) - 1] == *sep) {
		(void) snprintf(pat, sizeof(pat), "^%c([^%c]+)%c([^%c]*)%c$",
			*sep, *sep, *sep, *sep, *sep);
	} else {
		(void) snprintf(pat, sizeof(pat), "^%c([^%c]+)%c([^%c]*)$",
			*sep, *sep, *sep, *sep);
	}
	if (regcomp(&r, pat, REG_EXTENDED) != 0) {
		(void) fprintf(stderr, "bad regexec '%s'\n", pat);
		return 0;
	}
	if (regexec(&r, sep, 10, matchv, 0) != 0) {
		(void) fprintf(stderr, "no regexp match for '%s'\n", pat);
		return 0;
	}
	regfree(&r);
	fromlen = (int)(matchv[1].rm_eo - matchv[1].rm_so);
	if ((from = calloc(1, (unsigned)(fromlen + 1))) == NULL) {
		(void) fprintf(stderr, "bad calloc 1\n");
		return 0;
	}
	(void) memcpy(from, &sep[(int)matchv[1].rm_so], (unsigned)fromlen);
	tolen = (int)(matchv[2].rm_eo - matchv[2].rm_so);
	if ((to = calloc(1, (unsigned)(tolen + 1))) == NULL) {
		free(from);
		(void) fprintf(stderr, "bad calloc 2\n");
		return 0;
	}
	(void) memcpy(to, &sep[(int)matchv[2].rm_so], (unsigned)tolen);
	if (regcomp(&r, from, REG_EXTENDED) != 0) {
		free(from);
		free(to);
		(void) fprintf(stderr, "bad regexp '%s'\n", from);
		return 0;
	}
	for (inp = in, outp = out ;
	     (size_t)(inp - in) < insize && (size_t)(outp - out) < outsize;
	     ) {
		if (regexec(&r, inp, 10, matchv, 0) != 0) {
			break;
		}
		/* copy from inp to matchv[0].rm_so into outp and advance */
		(void) memcpy(outp, inp, (size_t)matchv[0].rm_so);
		inp += (int)matchv[0].rm_eo;
		outp += (int)matchv[0].rm_eo - 1;
		/* do substitution */
		(void) memcpy(outp, to, (unsigned)tolen);
		outp += tolen;
	}
	/* copy from inp to end into outp and advance */
	(void) memcpy(outp, inp, (size_t)(insize - (size_t)(inp - in)));
	outp += (size_t)(insize - (size_t)(inp - in));
	regfree(&r);
	free(from);
	free(to);
	return (int)(outp - out);
}

/* hand edit the input */
static int
edit(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	const char	*editor;
	const char	*tmpdir;
	char		*out = (char *)outvp;
	char		*argv[10];
	char		 tmppath[MAXPATHLEN];
	int		 status;
	int		 pid;
	int		 cc;
	int		 wc;
	int		 fd;

	USE_ARG(codecs);
	USE_ARG(op);
	if ((tmpdir = getenv("TMPDIR")) == NULL) {
		tmpdir = "/tmp";
	}
	(void) snprintf(tmppath, sizeof(tmppath), "%s/codecs.XXXXXX", tmpdir);
	if ((fd = mkstemp(tmppath)) < 0) {
		(void) fprintf(stderr, "bad temp file '%s'\n", tmppath);
		return 0;
	}
	for (cc = 0 ; (wc = write(fd, &in[cc], insize - cc)) > 0 ; cc += wc) {
	}
	if ((editor = getenv("VISUAL")) == NULL) {
		editor = getenv("EDITOR");
	}
	if (editor == NULL) {
		editor = "vi";
	}
	argv[0] = __UNCONST(editor);
	argv[1] = tmppath;
	argv[2] = NULL;
	switch(pid = fork()) {
	case -1:
		break;
	case 0:
		/* child */
		execv(editor, argv);
		break;
	default:
		while (wait(&status) != pid) {
		}
		break;
	}
	(void) lseek(fd, 0, SEEK_SET);
	for (cc = 0 ; (wc = read(fd, &out[cc], outsize - cc)) > 0 ; cc += wc) {
	}
	(void) unlink(tmppath);
	(void) close(fd);
	return cc;
}

/* convert a number of seconds in 'in' to a string */
static int
secs2str(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	struct tm	*tm;
	time_t		 t;
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(insize);
	USE_ARG(op);
	t = (time_t) strtoll(in, NULL, 10);
	tm = localtime(&t);
	return snprintf(out, outsize, "%4.4d:%2.2d:%2.2d:%2.2d:%2.2d:%2.2d\n",
		tm->tm_year + 1900,
		tm->tm_mon + 1,
		tm->tm_mday,
		tm->tm_hour,
		tm->tm_min,
		tm->tm_sec);
}

/* convert a string to a number of seconds */
static int
str2secs(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	regmatch_t	 matchv[10];
	struct tm	 tm;
	struct tm	*tp;
	regex_t		 r;
	time_t		 t;
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(insize);
	USE_ARG(op);
	t = time(NULL);
	tp = localtime(&t);
	(void) memcpy(&tm, tp, sizeof(tm));
	(void) regcomp(&r,
		"([0-9]+):([0-9]+):([0-9]+):([0-9]+):([0-9]+):([0-9]+)",
		REG_EXTENDED);
	if (regexec(&r, in, 10, matchv, 0) == 0) {
		tm.tm_year = re2num(in, matchv[1].rm_so, matchv[1].rm_eo) - 1900;
		tm.tm_mon = re2num(in, matchv[2].rm_so, matchv[2].rm_eo) - 1;
		tm.tm_mday = re2num(in, matchv[3].rm_so, matchv[3].rm_eo);
		tm.tm_hour = re2num(in, matchv[4].rm_so, matchv[4].rm_eo);
		tm.tm_min = re2num(in, matchv[5].rm_so, matchv[5].rm_eo);
		tm.tm_sec = re2num(in, matchv[6].rm_so, matchv[6].rm_eo);
	}
	t = mktime(&tm);
	return snprintf(out, outsize, "%lld\n", (long long)t);
}

/* randomise insize bytes in out */
static int
randomise(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	uint32_t	 r;
	size_t		 i;
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(in);
	USE_ARG(insize);
	USE_ARG(op);
	for (i = 0 ; i < outsize ; i++) {
		r = random();
		(void) memcpy(&out[i], &r, MIN((unsigned)(outsize - i), sizeof(r)));
	}
	return outsize;
}

/* rotate characters in the alphabet */
static int
rotate(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	size_t		 i;
	char		*out = (char *)outvp;
	int		 rot;

	USE_ARG(codecs);
	USE_ARG(outsize);
	rot = re2num(op, 3, strlen(op));
	for (i = 0 ; i < insize ; i++) {
		if (islower((uint8_t)in[i])) {
			out[i] = (((in[i] - 'a') + rot) % 26) + 'a';
		} else if (isupper((uint8_t)in[i])) {
			out[i] = (((in[i] - 'A') + rot) % 26) + 'A';
		} else {
			out[i] = in[i];
		}
	}
	return insize;
}

static const char	*rad50 = " ABCDEFGHIJKLMNOPQRSTUVWXYZ$.%0123456789";

/* DEC RADIX-50 encoding */
static int
rad50encode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	const char	*cp;
	uint16_t	 ch;
	size_t		 i;
	size_t		 o;
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(op);
	USE_ARG(outsize);
	for (i = 0, o = 0 ; i < insize ; i += 3, o += 2) {
		ch = 0;
		if ((cp = strchr(rad50, toupper((uint8_t)in[i]))) != NULL) {
			ch = (unsigned)(cp - rad50) * (40 * 40);
		}
		if ((cp = strchr(rad50, toupper((uint8_t)in[i + 1]))) != NULL) {
			ch += (unsigned)(cp - rad50) * 40;
		}
		if ((cp = strchr(rad50, toupper((uint8_t)in[i + 2]))) != NULL) {
			ch += (unsigned)(cp - rad50);
		}
		(void) memcpy(&out[o], &ch, sizeof(ch));
	}
	return (int)o;
}

/* DEC RADIX-50 decode */
static int
rad50decode(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	uint16_t	 ch;
	size_t		 i;
	size_t		 o;
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (i = 0, o = 0 ; i < insize && o < outsize ; i += 2, o += 3) {
		(void) memcpy(&ch, &in[i], sizeof(ch));
		out[o] = rad50[ch / (40 * 40)];
		ch %= (40 * 40);
		out[o + 1] = rad50[ch / 40];
		ch %= 40;
		out[o + 2] = rad50[ch];
	}
	return (int)o;
}

/* bin2hex encoding */
static int
bin2hex(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	size_t	 i;
	size_t	 o;
	char	*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (i = 0, o = 0 ; i < insize && o < outsize ; i++) {
		o += snprintf(&out[o], outsize - o, "\\x%.02x", (uint8_t)in[i]);
	}
	return (int)o;
}

/* hex2bin encoding */
static int
hex2bin(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	size_t	 i;
	size_t	 o;
	char	*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(op);
	for (i = 0, o = 0 ; i < insize && o < outsize ; i += 4, o++) {
		out[o] = (uint8_t)(hexchar2num(in[i + 2]) * 16) +
				hexchar2num(in[i + 3]);
	}
	return (int)o;
}

/* perform an MD5 digest */
static int
md5(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	USE_ARG(codecs);
	USE_ARG(op);
	USE_ARG(outsize);
	(void) MD5Data((const uint8_t *)in, insize, outvp);
	return MD5_DIGEST_STRING_LENGTH;
}

/* perform an RMD160 digest */
static int
rmd160(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	USE_ARG(codecs);
	USE_ARG(op);
	USE_ARG(outsize);
	(void) RMD160Data((const uint8_t *)in, insize, outvp);
	return RMD160_DIGEST_STRING_LENGTH;
}

/* perform an SHA1 digest */
static int
sha1(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	USE_ARG(codecs);
	USE_ARG(op);
	USE_ARG(outsize);
	(void) SHA1Data((const uint8_t *)in, insize, outvp);
	return SHA1_DIGEST_STRING_LENGTH;
}

/* perform an SHA256 digest */
static int
sha256(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	USE_ARG(codecs);
	USE_ARG(op);
	USE_ARG(outsize);
	(void) SHA256_Data((const uint8_t *)in, insize, outvp);
	return SHA256_DIGEST_STRING_LENGTH;
}

/* perform an SHA512 digest */
static int
sha512(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	USE_ARG(codecs);
	USE_ARG(op);
	USE_ARG(outsize);
	(void) SHA512_Data((const uint8_t *)in, insize, outvp);
	return SHA512_DIGEST_STRING_LENGTH;
}

/* find the size of the input */
static int
sizefunc(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	USE_ARG(codecs);
	USE_ARG(in);
	USE_ARG(op);
	return snprintf(outvp, outsize, "%lld\n", (long long) insize);
}

/* byte swap a 32-bit integer */
static int
byteswap32(const char *in, char *out)
{
	out[0] = in[3];
	out[1] = in[2];
	out[2] = in[1];
	out[3] = in[0];
	return 4;
}

/* byte swap a 64-bit integer */
static int
byteswap64(const char *in, char *out)
{
	out[0] = in[7];
	out[1] = in[6];
	out[2] = in[5];
	out[3] = in[4];
	out[4] = in[3];
	out[5] = in[2];
	out[6] = in[1];
	out[7] = in[0];
	return 8;
}

/* bswap* */
static int
byteswap(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(insize);
	USE_ARG(outsize);
	if (strcmp(op, "bswap16") == 0) {
		out[0] = in[1];
		out[1] = in[0];
		return 2;
	}
	if (strcmp(op, "bswap32") == 0) {
		return byteswap32(in, out);
	}
	if (strcmp(op, "bswap64") == 0) {
		return byteswap64(in, out);
	}
	return 0;
}

/* bswap* */
static int
endianness(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	int	indian = 1;

	USE_ARG(codecs);
	USE_ARG(in);
	USE_ARG(insize);
	USE_ARG(op);
	return snprintf(outvp, outsize, "%s\n", (*(char *)(void *)&indian) ?
		"1234" : "4321");
}

/* zero insize bytes in out */
static int
zero(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(in);
	USE_ARG(insize);
	USE_ARG(op);
	(void) memset(out, 0x0, outsize);
	return outsize;
}

/* zero insize bytes in out */
static int
geterrors(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	USE_ARG(in);
	USE_ARG(insize);
	USE_ARG(op);
	return snprintf(outvp, outsize, "%.*s", codecs->errc, codecs->errors);
}

enum {
	MAX_SOUNDEX	= 4
};

/* calculate the soundex component for the input string */
static int
soundex(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	size_t		 i;
	size_t		 o;
	char		 ch;
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(op);
	USE_ARG(outsize);
	out[0] = tolower((unsigned char)*in);
	for (o = 1, i = 1 ; i < insize && o < MAX_SOUNDEX ; i++) {
		switch(tolower((uint8_t)in[i])) {
		case 'b':
		case 'f':
		case 'p':
		case 'v':
			ch = '1';
			break;
		case 'c':
		case 'g':
		case 'j':
		case 'k':
		case 'q':
		case 's':
		case 'x':
		case 'z':
			ch = '2';
			break;
		case 'd':
		case 't':
			ch = '3';
			break;
		case 'l':
			ch = '4';
			break;
		case 'm':
		case 'n':
			ch = '5';
			break;
		case 'r':
			ch = '6';
			break;
		default:
			continue;
		}
		if (out[o - 1] != ch) {
			out[o++] = ch;
		}
	}
	for ( ; o < MAX_SOUNDEX ; o++) {
		out[o] = '0';
	}
	out[o++] = '\n';
	out[o++] = 0x0;
	return o;
}

/* the following metaphone code is from:
 *	http://aspell.net/metaphone/metaphone-kuhn.txt
 * as posted to Usenet By Michael Kuhn in 1995
 */

/*
consonants:
      B  X  S  K  J  T  F  H  L  M  N  P  R  0  W  Y

exceptions:
^ae
^gn
^kn
^pn
^wr	-> all drop first letter
^x	-> s
^wh	-> w

transforms:
b -> B (unless at end of word after 'm', in which case lose it)
c -> X (if cia or ch)
  -> S (if ci or ce or cy)
  -> drop if sci, sce or scy
  -> K otherwise
d -> J (if dge, dgy or dgi)
  -> T otherwise
f -> F
g -> drop if gh and not at end or before vowel
  -> drop if gn or gned
  -> drop if dge as above
  -> K otherwise
h -> drop if after vowel and no vowel follows
  -> drop if after ch, sh, ph, th or gh
  -> H otherwise
j -> J
k -> drop if after c
  -> K otherwise
l -> L
m -> M
n -> N
p -> F if in ph
  -> P otherwise
q -> K
r -> R
s -> X if before h, or in sio
  -> S otherise
t -> X if tia or tio
  -> 0 if before h
  -> drop if in tch
  -> T otherwise
v -> F
w -> drop if not followed by a vowel
  -> W otherwise
x -> KS
y -> drop if not followed by a vowel
  -> Y if followed by a vowel
z -> S
*/

typedef struct except_t {
	const char	*match;
	int		 len;
	const char	*xform;
} except_t;

static except_t	exceptions[] = {
	{	"ae",	2,	"Ie"	},
	{	"gn",	2,	"In"	},
	{	"kn",	2,	"In"	},
	{	"pn",	2,	"In"	},
	{	"wr",	2,	"Ir"	},
	{	"wh",	2,	"Iw"	},
	{	NULL,	0,	NULL	}
};

#define STRMATCH2(s1,ind,c1,c2)	\
	((s1)[ind] == (c1) && (s1)[ind + 1] == (c2))
#define STRMATCH3(s1,ind,c1,c2,c3)	\
	((s1)[ind] == (c1) && (s1)[ind + 1] == (c2) && (s1)[ind + 2] == (c3))

static int
metaphone(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	except_t	*ep;
	size_t		 chars;
	size_t		 i;
	size_t		 r;
	size_t		 o;
	char		*out = (char *)outvp;
	char		*buf;

	USE_ARG(codecs);
	USE_ARG(op);
	USE_ARG(outsize);
	/* copy to working storage, expanding for any 'x's */
	if ((buf = calloc(1, insize)) == NULL) {
		add_error(codecs, "metaphone: bad allocation %u bytes\n", insize);
		return 0;
	}
	for (r = o = i = 0 ; r < insize ; i++, o++, r++) {
		if ((buf[o] = buf[i] = tolower((uint8_t)in[i])) == 'x') {
			buf[i++] = buf[o++] = 'k';
			buf[i] = buf[o] = 's';
		}
	}
	buf[i] = 0x0;
	chars = i;
	/* beginning exceptions first */
	for (o = i = 0, ep = exceptions ; ep->len ; ep++) {
		if (strncmp(buf, ep->match, (unsigned)ep->len) == 0) {
			buf[o++] = ep->xform[i++];
			buf[o++] = ep->xform[i++];
		}
	}
	/* main transforms */
	for ( ; i < chars ; i++, o++) {
		switch(tolower((uint8_t)buf[i])) {
		case 'a':
		case 'e':
		case 'i':
		case 'o':
		case 'u':
			buf[o] = 'I';
			break;
		case 'b':
			buf[o] = (i == chars - 1) ? 'I' : 'B';
			break;
		case 'c':
			if (STRMATCH2(buf, i + 1, 'i', 'a')) {
				buf[o] = 'X';
			} else if (strchr("iey", buf[i + 1]) != NULL) {
				buf[o] = 'S';
			} else if (STRMATCH3(buf, i - 1, 's', 'c', 'e') ||
				   STRMATCH3(buf, i - 1, 's', 'c', 'i') ||
				   STRMATCH3(buf, i - 1, 's', 'c', 'y')) {
				buf[o] = 'I';
			} else {
				buf[o] = 'K';
			}
			break;
		case 'd':
			if (STRMATCH2(buf, i + 1, 'g', 'e') ||
			    STRMATCH2(buf, i + 1, 'g', 'y') ||
			    STRMATCH2(buf, i + 1, 'g', 'i')) {
				buf[o] = 'J';
				buf[o] = 'I';
				o += 1;
				i += 1;
			} else {
				buf[o] = 'T';
			}
			break;
		case 'f':
		case 'j':
		case 'l':
		case 'm':
		case 'n':
		case 'r':
			buf[o] = toupper((unsigned char)buf[i]);
			break;
		case 'g':
			if (buf[i + 1] == 'h' &&
			    (i < chars - 2 || strchr("aeiou", buf[i + 2]) == NULL)) {
				buf[o] = 'I';
			} else if (STRMATCH3(buf, i + 1, 'n', 'e', 'd')) {
				buf[o] = 'I';
			} else {
				buf[o] = 'K';
			}
			break;
		case 'h':
			if (strchr("aeiou", buf[i - 1]) != NULL && 
			    strchr("aeiou", buf[i + 1]) == NULL) {
				buf[o] = 'I';
			} else if (strchr("csptg", buf[i - 1]) != NULL) {
				buf[o] = 'I';
			} else {
				buf[o] = 'K';
			}
			break;
		case 'k':
			buf[o] = (buf[i - 1] == 'c') ? 'I' : 'K';
			break;
		case 'p':
			buf[o] = (buf[i + 1] == 'h') ? 'F' : 'P';
			break;
		case 'q':
			buf[o] = 'K';
			break;
		case 's':
			if (buf[i + 1] == 'h' || STRMATCH3(buf, i, 's', 'i', 'o')) {
				buf[o] = 'X';
			} else {
				buf[o] = 'S';
			}
			break;
		case 't':
			if (STRMATCH2(buf, i + 1, 'i', 'a') ||
			    STRMATCH2(buf, i + 1, 'i', 'o')) {
				buf[o] = 'X';
			} else if (buf[i + 1] == 'h') {
				buf[o] = '0';
			} else if (STRMATCH2(buf, i + 1, 'c', 'h')) {
				buf[o] = 'I';
			} else {
				buf[o] = 'T';
			}
			break;
		case 'v':
			buf[o] = 'F';
			break;
		case 'w':
			if (strchr("aeiou", buf[i + 1]) == NULL) {
				buf[o] = 'I';
			} else {
				buf[o] = 'W';
			}
			break;
		case 'y':
			if (strchr("aeiou", buf[i + 1]) == NULL) {
				buf[o] = 'I';
			} else {
				buf[o] = 'Y';
			}
			break;
		case 'z':
			buf[o] = 'S';
			break;
		}
	}
	/* copy to output buffer */
	for (o = 0, i = 0 ; i < chars ; i++) {
		if ((out[o] = buf[i]) != 'I' && out[o] != out[o - 1]) {
			o += 1;
		}
	}
	out[o] = 0x0;
	free(buf);
	return o;
}

/* return the IP address (v4 and v6 if available) for a host */
static int
resolve(codecs_t *codecs, const char *in, const size_t insize, const char *op,
	void *outvp, size_t outsize)
{
	struct sockaddr_in6	*sin6;
	struct sockaddr_in	*sin;
	struct addrinfo		*res0;
	struct addrinfo		*res;
	size_t			 o;
	char			 input[512];
	char			*out = (char *)outvp;
	int			 found;
	int			 done[7];
	int			 cc;

	USE_ARG(codecs);
	USE_ARG(insize);
	USE_ARG(op);
	if (snprintf(input, sizeof(input), "%.*s", (int)insize, in) >
			(int)sizeof(input)) {
		add_error(codecs, "resolve: ip addr too large '%s'\n", in);
		return 0;
	}
	if ((found = getaddrinfo(input, NULL, NULL, &res0)) != 0) {
		/* possibly trim off a trailing \n */
		if (input[insize - 1] == '\n') {
			input[insize - 1] = 0x0;
		}
		found = getaddrinfo(input, NULL, NULL, &res0);
	}
	if (found != 0) {
		add_error(codecs,
			"resolve: no address info for '%s'\n", in);
		return 0;
	}
	o = 0;
	done[4] = done[6] = 0;
	for (res = res0 ; res != NULL && o < outsize - 2 ; res = res->ai_next) {
		if (res->ai_family == AF_INET && !done[4]) {
			sin = (struct sockaddr_in *)(void *)res->ai_addr;
			cc = snprintf(&out[o], outsize - o, "%s ",
				inet_ntoa(sin->sin_addr));
			if (cc > (int)(outsize - o)) {
				return add_error(codecs,
					"resolve: address '%s' truncated\n", in);
			}
			o += cc;
			done[4] = 1;
		}
		if (res->ai_family == AF_INET6 && !done[6]) {
			sin6 = (struct sockaddr_in6 *)(void *)res->ai_addr;
			(void) getnameinfo((const struct sockaddr *)(void *)sin6,
				(unsigned)sin6->sin6_len,
				&out[o], outsize - o - 1,
				NULL, 0, NI_NUMERICHOST);
			o += strlen(&out[o]);
			out[o++] = ' ';
			done[6] = 1;
		}
	}
	freeaddrinfo(res0);
	if (o > outsize - 1) {
		o = outsize - 1;
	}
	out[o] = 0x0;
	return o;
}

/* return the hostname for an IP address (v4 or v6) */
static int
reverse_resolve(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	struct sockaddr_in6	*sin6;
	struct sockaddr_in	*sin;
	struct addrinfo		 hints;
	struct addrinfo		*res0;
	struct addrinfo		*res;
	size_t			 o;
	char			 input[512];
	char			*out = (char *)outvp;
	int			 found;

	USE_ARG(codecs);
	USE_ARG(insize);
	USE_ARG(op);
	(void) memset(&hints, 0x0, sizeof(hints));
	hints.ai_flags = AI_NUMERICHOST;
	if (snprintf(input, sizeof(input), "%.*s", (int)insize, in) >
			(int)sizeof(input)) {
		add_error(codecs, "reverse_resolve: addr too large '%s'\n", in);
		return 0;
	}
	if ((found = getaddrinfo(input, NULL, &hints, &res0)) != 0) {
		/* possibly trim off a trailing \n */
		if (input[insize - 1] == '\n') {
			input[insize - 1] = 0x0;
		}
		found = getaddrinfo(input, NULL, &hints, &res0);
	}
	if (found != 0) {
		add_error(codecs,
			"reverse_resolve: can't find host info for '%s'\n", in);
		return 0;
	}
	o = 0;
	for (res = res0 ; res && o < outsize - 2 ; res = res->ai_next) {
		if (res->ai_family == AF_INET) {
			sin = (struct sockaddr_in *)(void *)res->ai_addr;
			if (getnameinfo((const struct sockaddr *)(void *)sin,
					(unsigned)sin->sin_len,
					&out[o], outsize - o,
					NULL, 0, NI_NAMEREQD) == 0) {
				o = strlen(out);
				break;
			}
		}
		if (res->ai_family == AF_INET6) {
			sin6 = (struct sockaddr_in6 *)(void *)res->ai_addr;
			if (getnameinfo((const struct sockaddr *)(void *)sin6,
					(unsigned)sin6->sin6_len,
					&out[o], outsize - o,
					NULL, 0, NI_NAMEREQD) == 0) {
				o = strlen(out);
				break;
			}
		}
	}
	freeaddrinfo(res0);
	out[o] = 0x0;
	return o;
}

/* compress using zlib */
static int
zlib_compress(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	unsigned long	 size;
	uint64_t	 origsize;
	uint64_t	 netsize;
	char		*out = (char *)outvp;
	int		 indian = 1;

	USE_ARG(codecs);
	USE_ARG(op);
	size = outsize - sizeof(origsize);
	if (compress((uint8_t *)&out[sizeof(origsize)], &size,
		(const uint8_t *)in, (unsigned long)insize) != Z_OK) {
		return 0;
	}
	origsize = (uint64_t)insize;
	if (*(char *)(void *)&indian) {
		/* little-endian -- byteswap64 the size */
		(void) byteswap64((const char *)(const void *)&origsize,
				(char *)(void *)&netsize);
		origsize = netsize;
	}
	(void) memcpy(out, &origsize, sizeof(origsize));
	return size + sizeof(origsize);
}

/* decompress using zlib */
static int
zlib_decompress(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	unsigned long	 size;
	uint64_t	 origsize;
	uint64_t	 netsize;
	int		 indian = 1;

	USE_ARG(codecs);
	USE_ARG(op);
	size = outsize - sizeof(origsize);
	(void) memcpy(&origsize, in, sizeof(origsize));
	if (*(char *)(void *)&indian) {
		/* little-endian -- byteswap64 the size */
		(void) byteswap64((const char *)(const void *)&origsize,
				(char *)(void *)&netsize);
		origsize = netsize;
	}
	if (uncompress(outvp, &size, (const uint8_t *)&in[sizeof(origsize)],
		(unsigned long)(insize - sizeof(origsize))) != Z_OK) {
		return 0;
	}
	return (int)origsize;
}

/* compress using bzlib */
static int
bzlib_compress(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	const int	 blocksize = 9;
	const int	 verbose = 0;
	const int	 work = 30;
	unsigned	 size;
	char		*out = (char *)outvp;

	USE_ARG(codecs);
	USE_ARG(op);
	size = outsize;
	if (BZ2_bzBuffToBuffCompress(out, &size, __UNCONST(in), insize,
			blocksize, verbose, work) != BZ_OK) {
		return 0;
	}
	return size;
}

/* decompress using bzlib */
static int
bzlib_decompress(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	const int	verbose = 0;
	const int	small = 0;
	unsigned	size;

	USE_ARG(codecs);
	USE_ARG(op);
	size = outsize;
	if (BZ2_bzBuffToBuffDecompress(outvp, &size, __UNCONST(in), insize,
			verbose, small) != BZ_OK) {
		return 0;
	}
	return (int)size;
}

/* count character frequencies */
static int
frequencies(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	uint64_t	 freqs[256];
	size_t		 i;
	size_t		 o;
	size_t		 from;
	size_t		 to;
	char		*out = (char *)outvp;
	int		 rawfreqs;
	int		 cc;

	USE_ARG(codecs);
	rawfreqs = (strncmp(op, "raw", 3) == 0);
	for (i = 0 ; i < __arraycount(freqs) ; i++) {
		freqs[i] = 0;
	}
	for (i = 0 ; i < insize ; i++) {
		if (rawfreqs) {
			freqs[(uint8_t)in[i]] += 1;
		} else {
			freqs[tolower((uint8_t)in[i])] += 1;
		}
	}
	if (rawfreqs) {
		from = 0;
		to = __arraycount(freqs);
	} else {
		from = 'a';
		to = 'z' + 1;
	}
	for (i = from, o = 0 ; i < to ; i++) {
		cc = snprintf(&out[o], outsize - o, "%d %c\t%" PRIu64 "\n",
			i, (isprint((uint8_t)i)) ? i : '.', freqs[i]);
		if (cc > (int)(outsize - o) || cc < 0) {
			break;
		}
		o += cc;
	}
	return (int)o;
}

/* listcodecs -- overload op for codecs */
static int
listcodecs(codecs_t *codecs, const char *in, const size_t insize,
	const char *op, void *outvp, size_t outsize)
{
	unsigned	 i;
	codec_t		*codec;
	size_t		 o;
	char		*out = (char *)outvp;
	int		 cc;

	USE_ARG(in);
	USE_ARG(insize);
	USE_ARG(op);
	for (i = 0, o = 0, codec = codecs->v ;
	     o < outsize && i < codecs->c ;
	     o += cc, i++, codec++) {
		cc = snprintf(&out[o], outsize - o, "%s\n", codec->op);
		if (cc > (int)(outsize - o) || cc < 0) {
			break;
		}
	}
	return (int)o;
}

/* support functions from here */

/* return the value which is guaranteed to hold the output */
static int
multiplier(codecs_t *codecs, const char *mult, size_t insize)
{
	static regex_t	r;
	static int	compiled;
	regmatch_t	matchv[10];

	if (!compiled) {
		compiled = 1;
		(void) regcomp(&r, "^([0-9]+)%([-+][0-9]+)?", REG_EXTENDED);
	}
	if (regexec(&r, mult, 10, matchv, 0) != 0) {
		add_error(codecs,
			"multiplier: bad multiplier regex '%s'\n", multiplier);
		return 0;
	}
	return ((re2num(mult, matchv[1].rm_so, matchv[1].rm_eo) * insize) / 100) +
		re2num(mult, matchv[2].rm_so, matchv[2].rm_eo);
}

/* find the operation to perform */
static codec_t *
findop(codecs_t *codecs, const char *op, const unsigned matchc,
	regmatch_t *matchv)
{
	unsigned	 i;
	codec_t		*codec;

	for (codec = codecs->v, i = 0 ; op && i < codecs->c ; i++, codec++) {
		if (regexec(codec->r, op, matchc, matchv, 0) == 0) {
			return codec;
		}
	}
	return NULL;
}

/* add format-related transformations */
static void
codecs_add_format(codecs_t *codecs)
{
	const bool	needinput = true;
	const bool	noinput = false;

	codecs_add(codecs, "^asa", asa, "110%", needinput);
	codecs_add(codecs, "^bswap(16|32|64)", byteswap, "0%+8", needinput);
	codecs_add(codecs, "^(dos( to |2)unix)", dos2unix, "50%", needinput);
	codecs_add(codecs, "^endian", endianness, "0%+8", noinput);
	codecs_add(codecs, "^hexdump", hexdump, "500%+80", needinput);
	codecs_add(codecs, "^rot[0-9]+", rotate, "100%", needinput);
	codecs_add(codecs, "^secs2str", secs2str, "100%+30", needinput);
	codecs_add(codecs, "^str2secs", str2secs, "100%+30", needinput);
	codecs_add(codecs, "^unhexdump", unhexdump, "25%", needinput);
}

/* add charset-related transformations */
static void
codecs_add_charset(codecs_t *codecs)
{
	const bool	needinput = true;

	codecs_add(codecs, "^(atob|b(ase)?64[ ]*decode|from([-_ ])?base64)",
			xlate_b64decode, "80%", needinput);
	codecs_add(codecs, "^(btoa|b(ase)?64[ ]*encode|to([-_ ])?base64)",
			xlate_b64encode, "150%", needinput);
	codecs_add(codecs, "^b(ase)?85decode", base85decode, "90%",
			needinput);
	codecs_add(codecs, "^b(ase)?85encode", base85encode, "150%",
			needinput);
	codecs_add(codecs, "^bin2hex", bin2hex, "400%+10", needinput);
	codecs_add(codecs, "^(capitali[sz]e|(to([-_ ])?|2)upper)",
			mixed2upper, "100%", needinput);
	codecs_add(codecs, "^from([-_ ])?(uri|html)", uri2ascii, "100%",
			needinput);
	codecs_add(codecs, "^full(([-_ ])?)?uuencode", full_uuencode,
			"150%+30", needinput);
	codecs_add(codecs, "^hex2bin", hex2bin, "25%+10", needinput);
	codecs_add(codecs, "^metaphone", metaphone, "100%", needinput);
	codecs_add(codecs, "^rad50decode", rad50decode, "140%", needinput);
	codecs_add(codecs, "^rad50encode", rad50encode, "80%", needinput);
	codecs_add(codecs, "^soundex", soundex, "0%+6", needinput);
	codecs_add(codecs, "^strvis([-_ ])?c", xlate_strvis_c, "300%",
			needinput);
	codecs_add(codecs, "^strunvis([-_ ])?c", xlate_strunvis_c, "33%",
			needinput);
	codecs_add(codecs, "^strunvis", xlate_strunvis_def, "33%",
			needinput);
	codecs_add(codecs, "^strvis", xlate_strvis_def, "300%", needinput);
	codecs_add(codecs, "^(to([-_ ])?|2)(html|uri)", ascii2uri, "300%",
			needinput);
	codecs_add(codecs, "^(to([-_ ])?|2)lower", mixed2lower, "100%",
			needinput);
	codecs_add(codecs, "^(to([-_ ])?|2)upper", mixed2upper, "100%",
			needinput);
	codecs_add(codecs, "^(unix(to|2)dos)", unix2dos, "200%", needinput);
	codecs_add(codecs, "^uudecode", uudecode, "80%", needinput);
	codecs_add(codecs, "^uuencode", uuencode, "200%", needinput);
}

/* add compression-related transformations */
static void
codecs_add_compression(codecs_t *codecs)
{
	const bool	needinput = true;

	codecs_add(codecs, "^bzcompress|bzip|bzip2", bzlib_compress, "150%+8",
			needinput);
	codecs_add(codecs, "^bzdecompress|bunzip|bunzip2", bzlib_decompress,
			"900%", needinput);
	codecs_add(codecs, "^compress|gzip", zlib_compress, "110%+8",
			needinput);
	codecs_add(codecs, "^decompress|gunzip", zlib_decompress, "900%",
			needinput);
}

/* add edit-related transformations */
static void
codecs_add_edit(codecs_t *codecs)
{
	const bool	needinput = true;

	codecs_add(codecs, "^edit",  edit,   "300%", needinput);
	codecs_add(codecs, "^sed", sed, "300%", needinput);
	codecs_add(codecs, "^substring:([-]?[0-9]+):([-]?[0-9]+)",
			substring, "100%", needinput);
}

/* add network-related transformations */
static void
codecs_add_network(codecs_t *codecs)
{
	const bool	needinput = true;

	codecs_add(codecs, "^((get)?ipaddress|resolve)", resolve, "0%+1024",
			needinput);
	codecs_add(codecs, "^(gethost(info|name)|rev[-_]?resolve|reverse[-_]?resolve)",
			reverse_resolve, "0%+1024", needinput);
}

/* add digest-related transformations */
static void
codecs_add_digest(codecs_t *codecs)
{
	const bool	needinput = true;

	codecs_add(codecs, "^md5", md5, "0%+33", needinput);
	codecs_add(codecs, "^(rmd160|ripemd160)", rmd160, "0%+41", needinput);
	codecs_add(codecs, "^sha1", sha1, "0%+41", needinput);
	codecs_add(codecs, "^sha256", sha256, "0%+65", needinput);
	codecs_add(codecs, "^sha512", sha512, "0%+129", needinput);
	codecs_add(codecs, "^size", sizefunc, "0%+129", needinput);
}

/* add fill-related transformations */
static void
codecs_add_fill(codecs_t *codecs)
{
	const bool	noinput = false;

	codecs_add(codecs, "^(list|help|codecs|ls)", listcodecs, "100%",
			noinput);
	codecs_add(codecs, "^randomi[sz]e", randomise, "100%", noinput);
	codecs_add(codecs, "^zero", zero, "100%", noinput);
	codecs_add(codecs, "^geterrors", geterrors, "0%+1024", noinput);
}

/* add hash-related transformations */
static void
codecs_add_hash(codecs_t *codecs)
{
	const bool	needinput = true;

	codecs_add(codecs, "^(dumb|dumbmul|lennart|crc|perl|perlxor|python|"
		"mouse|bernstein|honeyman|pjw|bob|torek|byacc|tcl|gawk|"
		"gcc3_|gcc3_2_|nem)hash", hashfunc, "0%+4", needinput);
}

/* add analysis-related transformations */
static void
codecs_add_analysis(codecs_t *codecs)
{
	const bool	needinput = true;

	codecs_add(codecs, "^(raw)?freq(uencies|s)?", frequencies,
		"0%+6000", needinput);
}



/****************************************************************************/
/* exported functions start here */
/****************************************************************************/

/* return the number of bytes to hold the output */
int
codecs_size(codecs_t *codecs, const char *op, const size_t insize)
{
	regmatch_t	 matchv[10];
	codec_t		*xp;

	if (op == NULL) {
		return add_error(codecs, "codecs_size: no operation\n");
	}
	if (codecs->size == 0) {
		codecs_begin(codecs, "all", NULL);
	}
	if ((xp = findop(codecs, op, 10, matchv)) == NULL) {
		return add_error(codecs, "codecs_size: no translation '%s' found\n", op);
	}
	return multiplier(codecs, xp->multiplier, insize);
}

/* return an indication of whether an input buffer is needed */
bool
codecs_input_needed(codecs_t *codecs, const char *op)
{
	regmatch_t	 matchv[10];
	codec_t		*xp;

	if (op == NULL) {
		add_error(codecs,
			"codecs_input_needed: no operation\n");
		return false;
	}
	if (codecs->size == 0) {
		codecs_begin(codecs, "all", NULL);
	}
	if ((xp = findop(codecs, op, 10, matchv)) == NULL) {
		add_error(codecs, 
			"codecs_input_needed: no translation '%s' found\n", op);
		return false;
	}
	return xp->inputneeded;
}

/* translate input to output - separate arrays */
int
codecs_transform(codecs_t *codecs, const char *in, const size_t len, const char *op,
	void *vp, size_t size)
{
	regmatch_t	 matchv[10];
	codec_t		*xp;
	char		*out = (char *)vp;

	if (codecs->size == 0) {
		codecs_begin(codecs, "all", NULL);
	}
	if (op == NULL) {
		return add_error(codecs, "codecs_transform: no operation\n");
	}
	if ((xp = findop(codecs, op, 10, matchv)) == NULL) {
		return add_error(codecs,
			"codecs_transform: no translation '%s' found\n", op);
	}
	if (in == NULL && xp->inputneeded) {
		add_error(codecs, "codecs_transform: NULL input buffer\n");
		return 0;
	}
	return (*xp->func)(codecs, in, len, op, out, size);
}

/* translate input to output - dynamically-sized output array */
int
codecs_alloc_transform(codecs_t *codecs, const char *in, const size_t len, const char *op,
		void **vp, size_t *size)
{
	regmatch_t	 matchv[10];
	codec_t		*xp;
	char		**dyn = (char **)vp;

	if (codecs->size == 0) {
		codecs_begin(codecs, "all", NULL);
	}
	if (op == NULL) {
		add_error(codecs, "codecs_alloc_transform: no operation\n");
		return -1;
	}
	if ((xp = findop(codecs, op, 10, matchv)) == NULL) {
		add_error(codecs,
			"codecs_alloc_transform: no translation '%s' found\n", op);
		return -1;
	}
	if (in == NULL && xp->inputneeded) {
		add_error(codecs, "codecs_alloc_transform: NULL input buffer\n");
		return -1;
	}
	*size = multiplier(codecs, xp->multiplier, len);
	if ((*dyn = calloc(1, *size)) == NULL) {
		add_error(codecs,
			"codecs_alloc_transform: bad allocation (%u bytes)\n", *size);
		return -1;
	}
	return (*xp->func)(codecs, in, len, op, *dyn, *size);
}

/* translate input to output - in place */
int
codecs_inplace_transform(codecs_t *codecs, void *vp, int len, const char *op)
{
	size_t	 outsize;
	void	*genericout;
	char	*out;
	char	*in = (char *)vp;
	int	 ret;

	if (in == NULL) {
		return add_error(codecs,
			"codecs_inplace_transform: NULL input buffer\n");
	}
	if (codecs->size == 0) {
		codecs_begin(codecs, "all", NULL);
	}
	if ((ret = codecs_alloc_transform(codecs, in, (const unsigned)len,
			op, &genericout, &outsize)) < 0) {
		return -1;
	}
	(void) memcpy(in, out = genericout, (size_t)ret);
	in[ret] = 0x0;
	free(out);
	return ret;
}

/* return non-zero if this is a valid operation */
bool
codecs_valid_op(codecs_t *codecs, const char *op)
{
	regmatch_t	 matchv[10];

	if (codecs->size == 0) {
		codecs_begin(codecs, "all", NULL);
	}
	return (findop(codecs, op, 10, matchv) == NULL) ? false : true;
}

/* add the function to the table */
int
codecs_add(codecs_t *codecs, const char *op,
	int (*func)(codecs_t *, const char *, const size_t, const char *, void *, size_t),
	const char *mult, const bool inputneeded)
{
	unsigned	 newsize;
	codec_t		*codec;

	if (codecs->locked) {
		return add_error(codecs, "codecs_add: table is locked\n");
	}
	newsize = 0;
	if (codecs->c == 0) {
		newsize = 31;
		codec = calloc(newsize, sizeof(*codecs->v));
	} else if (codecs->c == codecs->size) {
		newsize = codecs->size + 31;
		codec = realloc(codecs->v,
				newsize * sizeof(*codecs->v));
		(void) memset(&codec[codecs->size],
				0x0,
				31 * sizeof(*codecs->v));
	}
	if (newsize) {
		/* we've done some new allocation */
		if (codec == NULL) {
			return add_error(codecs,
				"codecs_add: can't grow codec array for '%s'\n", op);
		}
		codecs->size = newsize;
		codecs->v = codec;
	}
	codec = &codecs->v[codecs->c++];
	if ((codec->op = strdup(op)) == NULL ||
	    (codec->r = calloc(1, sizeof(*codec->r))) == NULL ||
	    (codec->multiplier = strdup(mult)) == NULL) {
		return add_error(codecs,
			"codecs_add: can't allocate codec entry for '%s'\n", op);
	}
	if (regcomp(codec->r, op, REG_EXTENDED) != 0) {
		return add_error(codecs,
			"codecs_add: can't compile regexp for '%s'\n", op);
	}
	codec->func = func;
	codec->inputneeded = inputneeded;
	return 1;
}

/* load the table */
int
codecs_begin(codecs_t *codecs, const char *subset, ...)
{
	va_list	args;

	if (codecs->locked) {
		return add_error(codecs, "codecs_begin: codecs table is locked\n");
	}
	va_start(args, subset);
	while (subset != NULL) {
		if (strcmp(subset, "analysis") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_analysis(codecs);
		}
		if (strcmp(subset, "charset") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_charset(codecs);
		}
		if (strcmp(subset, "compression") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_compression(codecs);
		}
		if (strcmp(subset, "digest") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_digest(codecs);
		}
		if (strcmp(subset, "fill") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_fill(codecs);
		}
		if (strcmp(subset, "format") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_format(codecs);
		}
		if (strcmp(subset, "edit") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_edit(codecs);
		}
		if (strcmp(subset, "hash") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_hash(codecs);
		}
		if (strcmp(subset, "network") == 0 ||
		    strcmp(subset, "all") == 0) {
			codecs_add_network(codecs);
		}
		subset = va_arg(args, char *);
	}
	va_end(args);
	return 1;
}

/* free storage related to codecs */
int
codecs_end(codecs_t *codecs)
{
	unsigned	 i;
	codec_t		*codec;

	for (i = 0, codec = codecs->v ; i < codecs->c ; codec++, i++) {
		if (codec->r) {
			regfree(codec->r);
			free(codec->r);
		}
		free(codec->op);
		free(codec->multiplier);
	}
	free(codecs->v);
	codecs->c = codecs->size = 0;
	codecs->v = NULL;
	return 1;
}

/* make sure no changes can happen to the table */
int
codecs_lockdown(codecs_t *codecs)
{
	time_t	t;

	if (codecs->locked) {
		return add_error(codecs, "codecs_lockdown: codecs table is locked\n");
	}
	t = time(NULL);
	codecs->locked = t;
	return 1;
}
